/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        Spy Headquaters Agent
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.meta.hq;

// FIPA OS imports
import fipaos.platform.df.DFListener;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.fipaman.DFAgentDescription;

import fipaos.util.DIAGNOSTICS;

// JAVA imports
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.LinkedList;
import java.util.Vector;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;

import java.text.Collator;

// MY imports

/**
 * This class is keeping the list of agents registered with particular DF
 * Initial agent list is passed to the constructor. The changes are performed
 * throught the DFListener interface.
 */


public class DFAgentList implements DFListener,TreeModel {

  private static final String ROOT = "Agent List";

  private SortedMap _platforms;
  private List      _listeners;

  public DFAgentList(DFAgentDescription[] agents)
  {
    _platforms= new TreeMap(Collator.getInstance());
    _listeners= new LinkedList();

    if(agents!=null)
      for(int i= 0;i<agents.length; i++)
        registered(agents[i]);

    DIAGNOSTICS.println("DFAgentList started", DIAGNOSTICS.LEVEL_MIN);
  }

  /**
   * Add the agent.
   */
  public void registered(DFAgentDescription dfad)
  {
    String name= dfad.getAgentName();

    int at= name.indexOf('@');     // Get the agent's name+platform name
    String agent= name.substring(0, at);
    String platform= name.substring(at);

    SortedMap agents;

    // Find the platform's hashtable
    agents= (SortedMap)_platforms.get(platform);
    if(agents==null)  // New platform -> create
    {
      agents= new TreeMap(Collator.getInstance());
      _platforms.put(platform, agents);
      insertNotify(new TreeModelEvent(this, new Object [] {ROOT}));
    }

    agents.put(agent, dfad.getAgentID());  // Add the agent to the list
    insertNotify(new TreeModelEvent(this, new Object [] {ROOT, platform}));
  }

  public void modified(DFAgentDescription dfad)
  {
    // Ignore - we don't care
  }

  public void deregistered(DFAgentDescription dfad)
  {
    String name= dfad.getAgentName();

    int at= name.indexOf('@');            // Get the agent's name+platform name
    String agent= name.substring(0, at);
    String platform= name.substring(at);

    TreeMap agents;

    agents= (TreeMap)_platforms.get(platform);
    if(agents!=null)            // If the platform exists
    {
      agents.remove(agent);     // Remove agent
      if(agents.size()==0)      // Remove empty platform
      {
        _platforms.remove(platform);
        removeNotify(new TreeModelEvent(this, new Object [] {ROOT}));
      }
      else
        removeNotify(new TreeModelEvent(this, new Object [] {ROOT, platform}));
    }
  }

  public Object getRoot()
  {
    return ROOT;
  }

  public Object getChild(Object parent, int index)
  {
    Set children;

    if(parent.equals(ROOT))
      children= _platforms.keySet();
    else
      children= ((SortedMap)_platforms.get(parent)).keySet();

    Object[] a= children.toArray();
    if(a.length>index)
      return a[index];
    else
      return null;
  }

  public int getChildCount(Object parent)
  {
    Set children;

    if(parent.equals(ROOT))
      children= _platforms.keySet();
    else
      children= ((SortedMap)_platforms.get(parent)).keySet();

    return children.size();
  }

  public boolean isLeaf(Object node)
  {
    String s= (String)node;
    if(s.equals(ROOT))
      return _platforms.size()==0;
    if(s.startsWith("@"))
      return false;
    else
      return true;
  }

  public void valueForPathChanged(TreePath path, Object newValue)
  {
    // Non editable -> ignore
  }

  public int getIndexOfChild(Object parent, Object _child)
  {
    Set children;
    String child= (String)_child;

    if(parent.equals(ROOT))
      children= _platforms.keySet();
    else
      children= ((SortedMap)_platforms.get(parent)).keySet();

    int index= -1;
    String s= "";
    Iterator i= children.iterator();
    while( i.hasNext() && !child.equalsIgnoreCase(s))
    {
      index++;
      s= (String)i.next();
    }

    return index;
  }

  public void addTreeModelListener(TreeModelListener l)
  {
    _listeners.add(l);
  }

  public void removeTreeModelListener(TreeModelListener l)
  {
    if(_listeners.contains(l))
      _listeners.remove(l);
  }

  public AgentID getAgentID(String name)
  {
    int at= name.indexOf('@');     // Get the agent's name+platform name
    String agent= name.substring(0, at);
    String platform= name.substring(at);

    SortedMap agents;

    // Find the platform's hashtable
    agents= (SortedMap)_platforms.get(platform);
    if(agents!=null)  // New platform -> create
    {
      AgentID a;
      a= (AgentID)agents.get(agent);
      if(a!=null)
        return (AgentID)a.clone();
    }
    return null;
  }

  /**
   * Returns the list of names currently available in the DF.
   */
  public List getAgentList()
  {
    List l= new Vector();

    Iterator pi= _platforms.keySet().iterator();
    while(pi.hasNext())
    {
      String pname= (String)pi.next();
      SortedMap p= (SortedMap)_platforms.get(pname);
      Iterator ai= p.keySet().iterator();
      while(ai.hasNext())
        l.add( ((String)ai.next())+pname );
    }

    return l;
  }

  private void insertNotify(TreeModelEvent e)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((TreeModelListener)i.next()).treeStructureChanged(e);
  }

  private void removeNotify(TreeModelEvent e)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((TreeModelListener)i.next()).treeStructureChanged(e);
  }
}