/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        Spy Headquaters Agent
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.meta.hq;

// FIPA-OS imports
import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.mts.Message;
import fipaos.util.DIAGNOSTICS;

// JAVA imports
import javax.swing.*;
import java.util.*;
import java.awt.event.*;

// MY imports
import bp.meta.ont.*;

public class HQ implements ViewListener {

  // Agent management popup meni strings
  private static final String popRECRUIT    = "Recruit...";
  private static final String popDISMISS    = "Dismiss";
  private static final String popSHOW       = "Show";
  private static final String popHIDE       = "Hide";
  private static final String popSTATISTICS = "Statistics...";

  private static final int MAX_MSG_COUNT = 1000;
  private static final int DELETE_STEP   =  500;

  private static final int NO_MESSAGE    = -1;

  // Agent that owns this HQ
  HQAgent _owner;

  // Status bar to show some messages
  JLabel  _statusbar;
  // The views interrested in captured messages
  List    _hqlisteners;

  // List of names of shown agents
  List    _shownagents;

  // List of
  Map     _spyedagents;

  // FilterCreatorwindow (for faster show)
  FilterCreator _fc;

  // MessageView (for faster show)
  MessageView   _mv;

  // Message counter
  int     _msgids;
  int     _msgofs;

  // Message list - used to check duplicate messages
  TreeMap _msgIDs;
  // Message list - used to search by msg index
  List   _msgIndex;

  public HQ(HQAgent owner)
  {
    _owner= owner;
    _statusbar= null;
    _hqlisteners= new Vector();

    _shownagents= new Vector();
    _spyedagents= new Hashtable();

    _fc= new FilterCreator(this);

    _mv= new MessageView(this);

    _msgids= 0;
    _msgofs= 0;
    _msgIDs= new TreeMap();
    _msgIndex= new Vector();
  }

  public synchronized void shutdown()
  {
    Iterator i= _spyedagents.values().iterator();
    while(i.hasNext())
      ((HQTask)i.next()).finish();

    _fc.dispose();
    _fc= null;

    _mv.dispose();
    _mv= null;
  }

  public void setStatusBar(JLabel statusbar)
  {
    _statusbar= statusbar;
  }

  public void recruitAgent(String agent)
  {
    if(!_spyedagents.containsKey(agent) && !_fc.isVisible())
      _fc.Execute(agent, _owner.getDFAgentList().getAgentList());
  }

  public void showAgent(String agent)
  {
    if(!_shownagents.contains(agent) && !_spyedagents.containsKey(agent))
    {
      _shownagents.add(agent);
      fireAddAgent(agent, false);
    }
  }

  public void dismissAgent(String agent)
  {
    HQTask t= (HQTask)_spyedagents.get(agent);
    if(t!=null)
    {
      t.finish();
//      _spyedagents.remove(agent);
//      fireAddAgent(agent, false);
    }
  }

  public void hideAgent(String agent)
  {
    _shownagents.remove(agent);

    HQTask t= (HQTask)_spyedagents.get(agent);
    if(t!=null)
    {
      t.finish();
//      _spyedagents.remove(agent);
    }
    else
      fireRemoveAgent(agent);
  }

  public synchronized void insertMessage(ACL acl, String s)
  {
    if(s==null) // No tag present - add one
    {
      s= "@@@"+acl.hashCode();
//      DIAGNOSTICS.println("Added new TAG: "+s, this, DIAGNOSTICS.LEVEL_MAX);
    }

    if(!_msgIDs.containsKey(s))
    {
      if(_msgIndex.size()>MAX_MSG_COUNT) // Throw out the old messages
      {
        DIAGNOSTICS.println("Cleanig up the message cache...", this, DIAGNOSTICS.LEVEL_MAX);
        for(int i= 0; i<DELETE_STEP; i++)
        {        // The pair (ID, ACL)
          EncapsACL a= (EncapsACL)_msgIndex.get(0);
          _msgIDs.remove(a._ID);
          _msgIndex.remove(0);
        }
        _msgofs= _msgofs+DELETE_STEP;
        DIAGNOSTICS.println("Clean up complete.", this, DIAGNOSTICS.LEVEL_MAX);
      }

      _msgIndex.add(new EncapsACL(s, acl));
      _msgIDs.put(s, acl);
      fireAddMessage(_msgids, acl);
      ++_msgids;
    }
  }

  public void addHQListener(HQListener l)
  {
    if(!_hqlisteners.contains(l))
    {
      l.addViewListener(this);
      _hqlisteners.add(l);
    }
  }

  public void removeHQListener(HQListener l)
  {
    if(_hqlisteners.contains(l))
    {
      l.removeViewListener(this);
      _hqlisteners.remove(l);
    }
  }

  public synchronized void startHQTask(String spy, MsgFilter filter)
  {
    _shownagents.add(spy);
    fireAddAgent(spy, false);
    _owner.newTask(new HQTask(this, _owner.getDFAgentList().getAgentID(spy), filter));
  }

  public synchronized void HQTaskStarted(HQTask t)
  {
    String n= t.getAgentName();
    if(!_spyedagents.containsKey(n))
      _spyedagents.put(n, t);
    if(!_shownagents.contains(n))
      _shownagents.add(n);
    fireAddAgent(n, true);  // Clear the SPY flag
  }

  public synchronized void HQTaskFinished(HQTask t)
  {
    String n= t.getAgentName();

    if(t.getFailure())
      writeStatusBar("Failed to spy agent: "+n); // @@THREAD
    else
      writeStatusBar("Spying finished: "+n);

    if(_spyedagents.containsKey(n))
      _spyedagents.remove(n);

    if(_shownagents.contains(n))
      fireAddAgent(n, false);
    else
      fireRemoveAgent(n);
  }

  public synchronized void showStatistics(String agent)
  {
    AgentID a= _owner.getDFAgentList().getAgentID(agent);
    if(a!=null)
    {
      StatsView _sv= new StatsView(_owner, a);
      _sv.show();
    }
    else
      writeStatusBar("Can't find agent's ID: "+agent);
  }

  /**
   * ViewListener methods/
   */

  public void AgentPopUp(MouseEvent e, final String agent)
  {
    JPopupMenu     popup;
    JMenuItem      menuItem;
    ActionListener listener;

    listener= new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String s= e.getActionCommand();
        if(s==popRECRUIT)
          recruitAgent(agent);
        else if(s==popDISMISS)
          dismissAgent(agent);
        else if(s==popSHOW)
          showAgent(agent);
        else if(s==popHIDE)
          hideAgent(agent);
        else if(s==popSTATISTICS)
          showStatistics(agent);
      }
    };

    popup = new JPopupMenu(agent);

    if(_spyedagents.containsKey(agent))
    {
      menuItem = new JMenuItem(popDISMISS);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
      menuItem = new JMenuItem(popHIDE);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
    }
    else if(_shownagents.contains(agent))  // SHOWN
    {
      menuItem = new JMenuItem(popRECRUIT);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
      menuItem = new JMenuItem(popHIDE);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
    }
    else             // not SHOWN
    {
      menuItem = new JMenuItem(popRECRUIT);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
      menuItem = new JMenuItem(popSHOW);
      menuItem.addActionListener(listener);
      popup.add(menuItem);
    }

    popup.addSeparator();
    menuItem = new JMenuItem(popSTATISTICS);
    menuItem.addActionListener(listener);
    popup.add(menuItem);

    //menuItem.addActionListener(this);

    popup.show(e.getComponent(), e.getX(), e.getY());
  }

  public void AgentClicked(String agent)
  {
    if(!_mv.isVisible())
      fireHighlightMessage(NO_MESSAGE);
    writeStatusBar(agent);
  }

  public void MsgClicked(int ID)
  {
    if((ID-_msgofs)>=0)
      writeStatusBar((((EncapsACL)_msgIndex.get(ID-_msgofs))._msg).getPerformative());
    else
      writeStatusBar("Message too old: information disposed.");

    if(_mv.isVisible())
      _mv.setMessage(ID);
    else
      fireHighlightMessage(ID);
  }

  public void MsgDoubleClicked(int ID)
  {
    int i= ID-_msgofs;

    if(!_mv.isVisible() && (i>=0) && (i<_msgIndex.size()))
      _mv.setMessage(ID);
  }

  public synchronized ACL getMessage(int ID)
  {
    int i= ID-_msgofs;

    if( (i>=0) && (i<_msgIndex.size()))
    {
      fireHighlightMessage(ID);
      ACL acl= ((EncapsACL)_msgIndex.get(ID-_msgofs))._msg;
      writeStatusBar(acl.getPerformative());
      return acl;
    }
    else
      return null;
  }

  public boolean hasNext(int ID)
  {
    return (ID-_msgofs+1)<_msgIndex.size();
  }

  public boolean hasPrev(int ID)
  {
    return (ID-_msgofs-1)>=0;
  }

  public void deselectMessage()
  {
    fireHighlightMessage(NO_MESSAGE);
  }

  public AgentID getAID()
  {
    return _owner.getAID();
  }

  private void writeStatusBar(String s)
  {
    if(_statusbar!=null)
      _statusbar.setText(s);
  }

  private void fireAddAgent(String name, boolean spy)
  {
    Iterator i= _hqlisteners.iterator();
    while(i.hasNext())
      ((HQListener)i.next()).addAgent(name, spy);
  }

  private void fireRemoveAgent(String name)
  {
    Iterator i= _hqlisteners.iterator();
    while(i.hasNext())
      ((HQListener)i.next()).deleteAgent(name);
  }

  private void fireAddMessage(int ID, ACL msg)
  {
    Iterator i= _hqlisteners.iterator();
    while(i.hasNext())
    {
      ((HQListener)i.next()).addMessage(ID, msg);
    }
  }

  private void fireHighlightMessage(int ID)
  {
    Iterator i= _hqlisteners.iterator();
    while(i.hasNext())
      ((HQListener)i.next()).highlightMessage(ID);
  }

  private class EncapsACL
  {
    public String _ID;
    public ACL    _msg;
    public EncapsACL(String id, ACL msg)
      { _ID= id; _msg= msg; }
  }
}