// *****************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//    Gerstner Laboratory, Czech Technical University in Prague,
//    Karlovo namesti 13, 121 35 Prague 2, Czech Republic. All Gerstner
//    Laboratory Modifications are Copyright (C) 2001 Gerstner Laboratory.
//    All Rights Reserved.
//
// *****************************************************************************

/**
 * Title:        Spy Headquaters Agent
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.meta.hq;

//----------------------------------------------------------------------

// FIPA Imports
import fipaos.ont.fipa.*;

import fipaos.ont.fipa.fipaman.*;

import fipaos.agent.*;
import fipaos.agent.task.SynchronousTask;
import fipaos.agent.task.DFSearchTask;
import fipaos.agent.task.Task;

import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.platform.df.DFListenerTask;

import fipaos.util.DIAGNOSTICS;

// JAVA imports
import javax.swing.UIManager;

// MY imports


public class HQAgent extends FIPAOSAgent
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	// This is the type of the agent that we use when registering with the DF
	private static final String AGENT_TYPE	= "HQ-agent";

  private static final int DF_SEARCH_TIMEOUT = 30000;


	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

  private HQ _hq= null;

  private HQAgentGUI _gui= null;

  private DFAgentList _agentlist= null;

  private DFListenerTask _dflistenertask= null;

  private bp.pingpong.PingPongAgent _debugpingpong= null;

	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------

	/**
	 * An agent constructor takes three parameters - the location of the platform
	 * profile file (usually fipa-os\profiles\platform.profile), the agent name
	 * (e.g. df, ams), and the owner of the agent (e.g. Nortel, Bob).
	 *
	 * @param platform_profile The location of the platform profile
	 * @param name The agent name
	 * @param owner The owner of this agent
	 */
	public HQAgent( String platform_profile, String name, String owner )
	{

	/* First you must initialise the agent with a call to the FIPAOSAgent
	 * constructor. The arguments are the same three arguments to this constructor
	 * plus some instantiation flags.
	 * Arg 4 specifies if the agent should use the Task Manager - this does not have to be true
	 * but if you want to use the automatic platform registration (as this agent does) then
	 * you should set this to true. Tasks also make agent construction easier for more
	 * complicated agents. A discussion of programming agents using states is in the
	 * tutorial but is not included in this agent.
	 */
		super( platform_profile, name, owner, true );

		// This call should always be made in the agent constructor before any messaging needs to occur.
		// It sets up the message buffers.
		startPushing();

		// Now we can register with the local AMS (i.e. the platform) as the fist action our agent
		// makes
		try
		{
			// Attempt to register with AMS
			registerWithAMS();
			DIAGNOSTICS.println("Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch (AMSRegistrationException amsre)
		{
			// An exception has occured - this indicates that the AMS registration failed for some reason
			// Display the exception
			DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );

			// We can easily find the exception reason from the exception....
			String reason = amsre.getExceptionReason();

			// If the exception reason is not "already-registered", we can probably ignore the failure
			// (since this Agent was probably not de-registered the last time it ran!)
			if ( reason == null || !reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED))
			{
				// Shutdown nicely :)
				shutdown();
				return;
			}

		}

		// Now we can register with the local DF
		try
		{
			// Register with DF
			registerWithDF( AGENT_TYPE );
			DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch (DFRegistrationException dfre)
		{
			// An exception has occured - this indicates that the DF registration failed for some reason
			// Display the exception
			DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );

			// We can easily find the exception reason from the exception....
			String reason = dfre.getExceptionReason();

			if ( reason == null || !reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) )
			{
				// Shutdown nicely :)
				shutdown();
				return;
			}

		} // end catch

    initHQ();

    initAgentList();

    initGUI();
	}


	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------

  private void initHQ()
  {
    _hq= new HQ(this);
  }

  /**
   * The _agentlist maintains the current list of DF registered agents (+DF+AMS)
   * Thhese agents are shown in JTree in the main window. Agent list is automatically
   * updated via the dfregistered-DFListenerTask
   */
  private void initAgentList()
  {
    // Create the DF Listener task, which updates DFAgent List
    // It might take a while, so print the message
    DIAGNOSTICS.println("Searching DF", this, DIAGNOSTICS.LEVEL_MAX);

    // Create the synchronous DFsearch task and get the result
    DFAgentDescription[] result= (DFAgentDescription[])SynchronousTask.executeTask(_tm,
                    new DFSearchTask(new DFAgentDescription(), true), DF_SEARCH_TIMEOUT);

    // Create the new agent list with the actual agents
    _agentlist= new DFAgentList(result);

    // DF+AMS are not included in DF search, so add them manually
    DFAgentDescription ams= new DFAgentDescription();
    ams.setAgentName(getLocalAMSAID());
    _agentlist.registered(ams);

    DFAgentDescription df= new DFAgentDescription();
    df.setAgentName(getLocalDFAID());
    _agentlist.registered(df);

    // Create the DFListenerTask for automatic updates of agent list
    _dflistenertask= new DFListenerTask(_agentlist, getLocalDFAID());
    _tm.newTask(_dflistenertask);

    DIAGNOSTICS.println("DF Search done.", this, DIAGNOSTICS.LEVEL_MAX);
  }

  private void initGUI()
  {
    boolean packFrame= false;

    // Create the GUI
    _gui = new HQAgentGUI(this, _hq);

    // Tell the HQ where to display messages
    _hq.setStatusBar(_gui.getStatusBar());

    // Add the Vizualizer
    _hq.addHQListener(_gui.getVisualizer());

    // Show the GUI
    _gui.setVisible(true);
  }


	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/**
	 * Shuts down the agent by first deregistering with the DF and AMS
	 * (if registered) and then invoking shutdown() on the FIPAOSAgent shell
	 */
	public synchronized void shutdown()
	{

    if(_dflistenertask.subscribedOK())
    {
      _dflistenertask.stopListening();
    }

    if(_hq!=null)
      _hq.shutdown();

		// Check if we've registered with the DF
		if (registeredWithDF() == true)
		{
			try
			{
				// Attempt to deregister with DF
				deregisterWithDF();
				DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
			}
			catch (DFRegistrationException dfre)
			{
				// A problem deregistering occured..... we can obtain the reason though!
				String reason = dfre.getExceptionReason();

				DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
			} // end catch
		}

		// Check if we're registered with the AMS
		if (registeredWithAMS() == true)
		{
			try
			{
				// Deregister with AMS
				deregisterWithAMS();
				DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
			}
			catch (AMSRegistrationException amsre)
			{
				// A problem deregistering occured..... we can obtain the reason though!
				String reason = amsre.getExceptionReason();

				DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
			}
		}

    // Close the window
    if(_gui!=null)
    {
      _gui.dispose();
      _gui= null;
    }

    if(_debugpingpong!=null)
      _debugpingpong.shutdown();

		// Now call shutdown in the agent shell to release the core components
		super.shutdown();
	}

  public DFAgentList getDFAgentList()
  {
    return _agentlist;
  }

  public void newTask(Task t)
  {
    _tm.newTask(t);
  }

  public void doneDFListenerTask(Task t)
  {
    // Empty - the method is here just to disable warning.
  }

  public void doneStatsTask(Task t)
  {
    // Empty - the method is here just to disable warning.
  }

  public void doneClearStatsTask(Task t)
  {
    // Empty - the method is here just to disable warning.
  }

  public void doneHQTask(Task t)
  {
    if(_hq!=null)
      _hq.HQTaskFinished((HQTask)t);
  }

	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------

  /**
   * A main method to enable the agent to be run from the command line.
   * The agent takes two command line parameters - the platform profile location,
   * which is set up by the SetupFIPAOS.bat file to be in the system variable %PLATFORM%,
   * and the agent name.
   */
	public static void main( String args[] )
	{

		// Check for the correct number of arguments
		if( args.length < 2 )
		{

			System.out.println( "Usage: fipaos.tutorial.HQAgent location-of-platform-profile agent-name" );

		}
		else
		{
      try {
        UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel" );
      }
      catch(Exception e) {
        e.printStackTrace();
      }

			// If number of args is correct then run the agent - in this case the ownership
			// is always "FIPA-OS"
			// args[ 0 ] = the platform profile location
			// args[ 1 ] = the agent name
			HQAgent a= new HQAgent( args[ 0 ], "SpyHQ", "MiDo" );
      //a._debugpingpong= new bp.pingpong.PingPongAgent( args[0], "PingPong0", "MiDo");
      //new fipaos.tool.IOTestAgent( args[0], "iot", "MiDo");
		}

	}


} // End HQAgent
