/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        Spy Headquaters Agent
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */


package bp.meta.hq;

// FIPA OS Imports
import fipaos.agent.task.Task;
import fipaos.agent.task.WaitTask;
import fipaos.agent.conversation.Conversation;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.mts.Message;
import fipaos.util.DIAGNOSTICS;

// Java imports

// My imports
import bp.meta.ont.Action;
import bp.meta.ont.MsgFilter;
import bp.meta.ont.SPYCONSTANTS;


public class HQTask extends Task {

  public static final int SUBSCRIBE_TIMEOUT = 10000;

  // Flag that the subscription was confirmed
  private boolean    _subscribed= false;

  // Indicates if the Task was shut down properly
  private boolean    _failure= true;

  private HQ         _hq;

  private AgentID    _owner;

  // The sniffed agent
  private AgentID    _spy;

  // The used filter
  private MsgFilter  _filter;

  // Subscribe-advertise conversation ID
  private String _conversationid;

  private WaitTask _wt;


  public HQTask(HQ hq, AgentID spy, MsgFilter filter)
  {
    _hq= hq;
    _owner= hq.getAID();
    _spy= spy;
    _filter= filter;
  }

  public void startTask()
  {
    Action a= new Action();

    _filter.setHQ(_owner);  //Just to ensure the HQ is set

    a.setAgent(_spy);
    a.setPredicate(SPYCONSTANTS.PUT);
    a.setSubject(_filter.toString());

    ACL acl= getNewConversation(FIPACONSTANTS.FIPA_SUBSCRIBE);

    acl.setSenderAID(_owner);
    acl.setReceiverAID(_spy);
    acl.setPerformative(FIPACONSTANTS.SUBSCRIBE);
    /**
     * acl.setContentObject(a); - doesn't work if a isn't serialisable
     */
    acl.setContentObject(a);
    acl.setLanguage(SPYCONSTANTS.SPY_LANGUAGE);
    acl.setOntology(SPYCONSTANTS.SPY_ORDER);

    forward(acl);

    _wt= new WaitTask(SUBSCRIBE_TIMEOUT);
    newTask( _wt );

    DIAGNOSTICS.println("HQTask started for "+_spy.getName(), this, DIAGNOSTICS.LEVEL_3);
  }

  public void handleInform(Conversation c)
  {
    if(!_subscribed)  // First inform confirms the subscription
    {
      _failure= false;
      _subscribed= true;
      _conversationid= c.getConversationID();
      _hq.HQTaskStarted(this);
      /** //@todo  Add the subscribee to the agent list */
    }
    else
    {
      ACL acl=  c.getACL(c.getLatestMessageIndex());
      if(acl.getOntology().equalsIgnoreCase(SPYCONSTANTS.SPY_REPORT))
        done();  // The target agent was shut down
      else
      {
        String s= acl.getInReplyTo(); // Here is hidden the unique ID
        Object o= acl.getContentObject();
        if(o instanceof ACL)
          _hq.insertMessage((ACL)o, s);
        else  // instanceof String
          try
          {
            _hq.insertMessage(new ACL((String)o), s);
          }
          catch (fipaos.parser.ParserException e)
          {
            DIAGNOSTICS.println("Unparsable message: "+(String)o, this, DIAGNOSTICS.LEVEL_MAX);
          }
      }
    }
  }

  public void handleNotUnderstood(Conversation c)
  {
    done();  //Some error ocured, can't subscribe
  }

  public void handleFailure(Conversation c)
  {
    if(_subscribed)
    {
      /** @todo Remove the agent from the list */
    }
    done();  //Spy stops the subscription
  }

  public void doneWaitTask(Task t)
  {
    _wt= null;
    if(!_subscribed)   // No response receiver, close the task
      done();
  }

  public void finish()
  {
    if(_subscribed)    // Send cancel
    {
      Action a= new Action();

      a.setAgent(_spy);
      a.setPredicate(SPYCONSTANTS.CLEAR);
      a.setSubject(SPYCONSTANTS.FILTER);

      ACL acl = new ACL();

      acl.setPerformative(FIPACONSTANTS.CANCEL);
      acl.setSenderAID(_owner);
      acl.setReceiverAID(_spy);
      acl.setContentObject(a.toString());
      acl.setLanguage(SPYCONSTANTS.SPY_LANGUAGE);
      acl.setOntology(SPYCONSTANTS.SPY_ORDER);
      acl.setProtocol(FIPACONSTANTS.FIPA_SUBSCRIBE);
      acl.setConversationID(_conversationid);

      forward(acl);

      /** @todo Remove Subscribee from the agent list */
    }
    done();
  }

  public String getAgentName()
  {
    return _spy.getName();
  }

  public boolean getFailure()
  {
    return _failure;
  }

  protected void done()
  {
    // Stop the wait task to awoit _tm warning
    if(_wt!=null)
      _tm.removeTask(_wt);

    DIAGNOSTICS.println("HQTask finished for "+_spy.getName(), this, DIAGNOSTICS.LEVEL_3);

    super.done();
  }
}