/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        Spy Headquaters Agent
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.meta.hq.visualizer;

// FIPA-OS imports
import fipaos.ont.fipa.ACL;
import fipaos.util.DIAGNOSTICS;

// JAVA imports
import javax.swing.*;
import java.util.*;
import java.awt.Color;
import java.awt.event.*;
import java.awt.Dimension;


// MY imports
import bp.meta.hq.HQListener;
import bp.meta.hq.ViewListener;


public class Visualizer implements HQListener {

  private Color _frontAgent= Color.yellow;
  private Color _backAgent= Color.blue;

  private JScrollPane      _pane;
  private AgentVisualizer  _avis;
  private MessageVisualizer _mvis;

  private Vector _agents;
  private Vector _msgs;

  private List   _listeners;

  public Visualizer()
  {
    _listeners= new LinkedList();

    _agents= new Vector();
    _msgs= new Vector();

    _avis= new AgentVisualizer(_agents);
    _mvis= new MessageVisualizer(_msgs);
    _pane= new JScrollPane(_mvis);
    _pane.setColumnHeaderView(_avis);
    _pane.setPreferredSize(new Dimension(VCONST.INIT_W, VCONST.INIT_H));

    _avis.addMouseListener(new AgentMouseListener());
    _mvis.addMouseListener(new MsgMouseListener());
  }

  public JScrollPane getScrollPane()
  {
    return _pane;
  }

  /**
   * Add new message to the view.
   */
  public void addMessage(int ID, ACL Message)
  {
    int pos;

    if(_msgs.size()==0)
      pos= 0;
    else
      pos= ((VisMessage)_msgs.lastElement()).getNextRow();
    _msgs.add(new VisMessage(ID, pos, _agents, Message));
    _mvis.setMsgCount(_msgs.size());
    _mvis.repaint();
  }

  /**
   * Highlight the messsage (after click, etc...)
   */
  public void highlightMessage(int ID)
  {
    _mvis.highlightMessage(ID);
  }

  /**
   * Add new agent. spy parameter sais if agent is sniffed.
   */
  public void addAgent(String name, boolean spy)
  {
    int index= _agents.indexOf(new VisAgent.StringComparator(name));

    if(index!=-1)  // Agent present, just change status
    {
      ((VisAgent)_agents.get(index))._spy= spy;
      _avis.repaint();
    }
    else
    {                      // +1 - 0 is "OTHERS"
      _agents.add(new VisAgent(_agents.size()+1, name, spy, _frontAgent, _backAgent));

      Iterator i= _msgs.iterator();     // Recreate the images
      while(i.hasNext())
        ((VisMessage)i.next()).createImage();

      setAgentCount(_agents.size());
      _pane.repaint();
    }
  }

  /**
   * Delete the agent.
   */
  public void deleteAgent(String name)
  {
    int index= _agents.indexOf(new VisAgent.StringComparator(name));

    if(index!=-1)
    {
      _agents.remove(index);
      Iterator i= _agents.iterator();
      int c= 1;
      while(i.hasNext())   // Recreate agent images
      {
        ((VisAgent)i.next()).createImage(c);
        c++;
      }

      i= _msgs.iterator();  // Recreate message images
      while(i.hasNext())
        ((VisMessage)i.next()).createImage();

      setAgentCount(_agents.size());
      _pane.repaint();
    }
  }

  /**
   * Add listener to be notifyed when View events occur.
   */
  public void addViewListener(ViewListener l)
  {
    _listeners.add(l);
  }

  /**
   * Remove the listener.
   */
  public void removeViewListener(ViewListener l)
  {
    _listeners.remove(l);
  }

  private void clearMessages()
  {
    _msgs.clear();
    _mvis.setMsgCount(_msgs.size());
    _mvis.repaint();
  }

  private void setAgentCount(int cnt)
  {
    _mvis.setAgentCount(_agents.size());
    _avis.setAgentCount(_agents.size());
  }

  private void fireAgentPopUp(MouseEvent e, String agent)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((ViewListener)i.next()).AgentPopUp(e, agent);
  }

  private void fireAgentClicked(String agent)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((ViewListener)i.next()).AgentClicked(agent);
  }

  private void fireMsgClicked(int ID)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((ViewListener)i.next()).MsgClicked(ID);
  }

  private void fireMsgDoubleclicked(int ID)
  {
    Iterator i= _listeners.iterator();
    while(i.hasNext())
      ((ViewListener)i.next()).MsgDoubleClicked(ID);
  }

  private class AgentMouseListener extends MouseAdapter
  {
    public AgentMouseListener() {}

    public void mousePressed(MouseEvent e) {
        int i= _agents.indexOf(new VisAgent.PointComparator(e.getPoint()));
        if(i!=-1)
          if(!mayBePopUp(e, i))
            fireAgentClicked(((VisAgent)_agents.get(i))._fullname);
    }

    public void mouseReleased(MouseEvent e) {
       int i= _agents.indexOf(new VisAgent.PointComparator(e.getPoint()));
       if(i!=-1)
          mayBePopUp(e, i);
    }

    private boolean mayBePopUp(MouseEvent e, int i)
    {
      if(e.isPopupTrigger())
      {
        fireAgentPopUp(e, ((VisAgent)_agents.get(i))._fullname );
        return true;
      }
      else
        return false;
    }
  }

  private class MsgMouseListener extends MouseAdapter
  {
    public MsgMouseListener() {}

    public void mouseClicked(MouseEvent e)
    {
      if((e.getModifiers() & e.BUTTON1_MASK)!=0)
      {
        int i= _msgs.indexOf(new VisMessage.PointComparator(e.getPoint()));
        if(i!=-1)
        {
          if(e.getClickCount()==2)
            fireMsgDoubleclicked(((VisMessage)_msgs.get(i))._ID);
        }
      }
    }

    public void mousePressed(MouseEvent e)
    {
      if(!mayBePopUp(e))
      {
        int i= _msgs.indexOf(new VisMessage.PointComparator(e.getPoint()));
        if(i!=-1)
          fireMsgClicked(((VisMessage)_msgs.get(i))._ID);
      }
    }

    public void mouseReleased(MouseEvent e)
    {
      mayBePopUp(e);
    }

    private boolean mayBePopUp(MouseEvent e)
    {
      if(e.isPopupTrigger())
      {
        JPopupMenu     popup= new JPopupMenu();
        JMenuItem      item= new JMenuItem("Clear Messages");

        item.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            clearMessages();
          }
        });

        popup.add(item);
        popup.show(e.getComponent(), e.getX(), e.getY());
        return true;
      }
      else
        return false;
    }
  }
}