/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        MessageSpy Ontology
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.meta.ont;


// FIPA OS imports
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.fipa.ACL;
import fipaos.mts.Message;
import fipaos.parser.sl.ParseException;
import fipaos.parser.ParserException;

// My imports
import bp.meta.util.StringPattern;
import bp.ont.Constructor;
import bp.ont.OntException;

// JAVA imports
import java.util.List;
import java.util.Iterator;


public class MsgFilter {

  /**
   * Agent ID of this filter Headquaters;
   */
  private AgentID _hq;

  // Conversation ID for this filter fipa-subscribe conversation
  private String  _conversationID;

  private String _logic;

  private StringPattern _performative;
  private StringPattern _sender;
  private StringPattern _receiver;
  private StringPattern _language;
  private StringPattern _ontology;
  private StringPattern _protocol;

  public MsgFilter(String Filter) throws OntException
  {
    _conversationID= null;

    try
    {
      Constructor p= new Constructor(Filter);  // Params

      _hq= new AgentID(p.GetStrParam(SPYCONSTANTS.HQ));

      // Check if we really have a filter term
      if(p.GetOp().equalsIgnoreCase(SPYCONSTANTS.FILTER))
      {
        _logic= p.GetStrParam(SPYCONSTANTS.LOGIC);
        if(!_logic.equalsIgnoreCase(SPYCONSTANTS.AND) && !_logic.equalsIgnoreCase(SPYCONSTANTS.OR))
          throw new OntException(_logic+ " is not a valid filter operator.");

        _performative= new StringPattern( p.GetStrParam(SPYCONSTANTS.PERFORMATIVE) );
        _sender= new StringPattern( p.GetStrParam(FIPACONSTANTS.SENDER) );
        _receiver= new StringPattern( p.GetStrParam(FIPACONSTANTS.RECEIVER) );
        _language= new StringPattern( p.GetStrParam(FIPACONSTANTS.LANGUAGE) );
        _ontology= new StringPattern( p.GetStrParam(FIPACONSTANTS.ONTOLOGY) );
        _protocol= new StringPattern( p.GetStrParam(FIPACONSTANTS.PROTOCOL) );
      }
      else
        throw  new OntException("Not a filter constructor: "+Filter);
    }
    catch(ParseException e)
    {
      throw new OntException("Error parsing MsgFilter: "+e.getMessage());
    }
    catch(ParserException e)
    {
      throw new OntException("Error parsing Headquaters AID: "+e.getMessage());
    }
    catch(OntException e)
    {
      throw e;
    }
    catch(Throwable e)
    {
      throw new OntException("Error while creating MsgFilter: "+Filter);
    }
  }

  public MsgFilter()
  {
    _hq= null;

    _logic= SPYCONSTANTS.AND;

    _performative= new StringPattern("");
    _sender= new StringPattern("");
    _receiver= new StringPattern("");
    _language= new StringPattern("");
    _ontology= new StringPattern("");
    _protocol= new StringPattern("");
  }

  public boolean Match(ACL a)
  {
//    ACL a= msg.getACL();

    // Default value for receiver is invariant to filter logic operation
    boolean receiver= _logic.equalsIgnoreCase(SPYCONSTANTS.AND);

    if(!_receiver.Empty())
    {
      // Check the receiver field - might be multiple receivers. At least
      // one of them should match with the filter
      List r= a.getReceiverAIDs();
      Iterator i= r.iterator();
      receiver= false;

      while( !receiver && i.hasNext() )
        receiver= _receiver.Match(((AgentID)i.next()).getName(), false);
    }

    if(_logic.equalsIgnoreCase(SPYCONSTANTS.OR))
      return receiver ||
             _performative.Match(a.getPerformative(), false) ||
             _sender.Match(a.getSenderAID().getName(), false) ||
             _language.Match(a.getLanguage(), false) ||
             _ontology.Match(a.getOntology(), false) ||
             _protocol.Match(a.getProtocol(), false);
    else
      return receiver &&
             _performative.Match(a.getPerformative(), true) &&
             _sender.Match(a.getSenderAID().getName(), true) &&
             _language.Match(a.getLanguage(), true) &&
             _ontology.Match(a.getOntology(), true) &&
             _protocol.Match(a.getProtocol(), true);;
  }

  public AgentID getHQ()
  {
    return _hq;
  }

  public void setHQ(AgentID HQ) { _hq= HQ; }

  public void setLogic(String Logic) throws OntException
  {
    if(!Logic.equalsIgnoreCase(SPYCONSTANTS.AND) && !Logic.equalsIgnoreCase(SPYCONSTANTS.OR))
        throw new OntException(Logic+ " is not a valid filter operator");
    _logic= Logic;
  }

  public void setReceiver(String Receiver)
    { _receiver = new StringPattern( Receiver );}
  public void setPerformative(String Performative)
    { _performative = new StringPattern( Performative );}
  public void setSender(String Sender)
    { _sender = new StringPattern( Sender );}
  public void setLanguage(String Language)
    { _language = new StringPattern( Language );}
  public void setOntology(String Ontology)
    { _ontology = new StringPattern( Ontology );}
  public void setProtocol(String Protocol)
    { _protocol = new StringPattern( Protocol );}
  public void setConversationID(String ID)
    { _conversationID= ID; }
  public String getConversationID()
    { return _conversationID; }

  public String toString()
  {
    String s;

    s=  "("+SPYCONSTANTS.FILTER+" ";
    if(_hq!=null)
      s= s+SPYCONSTANTS.HQ+" "+_hq+" ";

    s= s+SPYCONSTANTS.LOGIC+" "+_logic+" ";

    if(!_performative.Empty())
      s= s+SPYCONSTANTS.PERFORMATIVE+" "+_performative+" ";

    if(!_sender.Empty())
      s= s+FIPACONSTANTS.SENDER+" "+_sender+" ";

    if(!_receiver.Empty())
      s= s+FIPACONSTANTS.RECEIVER+" "+_receiver+" ";

    if(!_language.Empty())
      s= s+FIPACONSTANTS.LANGUAGE+" "+_language+" ";

    if(!_ontology.Empty())
      s= s+FIPACONSTANTS.ONTOLOGY+" "+_ontology+" ";

    if(!_protocol.Empty())
      s= s+FIPACONSTANTS.PROTOCOL+" "+_protocol;

    s= s+")";

    return s;
  }

  public static class ConversationComparator
  {
    private String _cID;

    public ConversationComparator(String cid)
      { _cID= cid; }

    public boolean equals(Object other)
    {
      return ((MsgFilter)other)._conversationID.equals(_cID);
    }
  }

  public static class HQComparator
  {
    private String _hq;

    public HQComparator(String hq)
      { _hq= hq; }

    public boolean equals(Object other)
    {
      return ((MsgFilter)other)._hq.getName().equalsIgnoreCase(_hq);
    }
  }
}