/******************************************************************************
*   Open Source Copyright Notice and License: MessageSpy for FIPA-OS
*
* Copyright notice:
*    This file is Copyright (C) 2001 Gerstner Laboratory, Czech Technical
*    University in Prague (CTU), Karlovo namesti 13, 121 35 Prague 2,
*    Czech Republic. All rights reserved.
*
* The contents of this file are subject to the Emorphia Public License
* (the "License"); you may not use this file except in compliance with the
* License.  You may obtain a copy of the License at http://www.emorphia.com/EPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
* for the specific language governing rights and limitations under the License.
*
* The Original Code is MessageSpy for FIPA-OS (by Gerstner Laboratory).
*
* The Initial Developer of the Original Code is Gerstner Laboratory.
*
* Contributor(s):
*   Michal Dobisek, Gerstner Laboratory
*
******************************************************************************/

/**
 * Title:        The MessageSpy Class
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */


package bp.meta.spy;


// Java imports
import java.util.*;

// FIPA OS imports
import fipaos.mts.MessageReceiver;
import fipaos.mts.Message;
import fipaos.mts.MessageSender;
import fipaos.agent.FIPAOSAgent;
import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.FIPACONSTANTS;

import fipaos.util.DIAGNOSTICS;

// My imports
import bp.meta.ont.*;
import bp.meta.ont.Action;
import bp.ont.OntException;


public class MessageSpy implements MessageReceiver,MessageSender {

  public MessageSpy(FIPAOSAgent Owner)
  {
    // Start counting at current time
    _mID= (new Date()).getTime();

    _owner= Owner;

    _instats= new MsgStats();
    _outstats= new MsgStats();

    // Filter to recognize the message directed to the SPY itself
    _internalf= new MsgFilter();
    _internalf.setHQ(_owner.getAID());
    _internalf.setReceiver(_owner.getAID().getName());
    _internalf.setLanguage(SPYCONSTANTS.SPY_LANGUAGE);
    _internalf.setOntology(SPYCONSTANTS.SPY_ORDER);

    _filter= new Vector();


    DIAGNOSTICS.println("SPY:Spy started.", _owner, DIAGNOSTICS.LEVEL_MAX);
  }

  /**
   * Implementation of fipaos.mts.MessageReceiver.
   * Processes the message and passes it ahead.
   */
  public void receiveMessage(Message msg)
  {
    /**@todo: Implement internal message processing.*/
    if(ProcessMessage(msg))
    {
      _instats.insert(msg.getACL().getPerformative());
      if(_receiver!=null)
        _receiver.receiveMessage(msg);
    }
  }

  /**
   * Implementation of fipaos.mts.MessageReceiver.
   */
  public void setMessageSender(MessageSender Sender)
  {
    _sender= Sender;
  }

  /**
   * Implementation of fipaos.mts.MessageSender.
   * Processes the message and passes it ahead.
   */
  public void sendMessage( Message msg )
  {
    /**@todo: Implement internal message processing.*/
    ++_mID;
    msg.getEnvelope().setSlotValue(SPYCONSTANTS.SPY_TAG, _owner.getAID().getName()+_mID);
    if(ProcessMessage(msg))
    {
      _outstats.insert(msg.getACL().getPerformative());
      if(_sender!=null)
        _sender.sendMessage(msg);
    }
  }


  /**
   * Implementation of fipaos.mts.MessageSender.
   */
	public void setMessageReceiver( MessageReceiver mr ) {
    _receiver= mr;
  }

  /**
   * Implementation of fipaos.mts.MessageSender.
   */
	public void shutdown() {
    /**@todo: Implement internal shut down*/
    Iterator i= _filter.iterator();
    while(i.hasNext())
    {
      MsgFilter mf= (MsgFilter)i.next();

      ACL cc= new ACL();

      DIAGNOSTICS.println("SPY:Sending cancel to "+mf.getHQ().getName(), _owner, DIAGNOSTICS.LEVEL_3);

      cc.setPerformative(FIPACONSTANTS.INFORM);
      cc.setSenderAID(_owner.getAID());
      cc.setReceiverAID(mf.getHQ());
      cc.setContentObject("(shutdown)");
      cc.setLanguage(SPYCONSTANTS.SPY_LANGUAGE);
      cc.setOntology(SPYCONSTANTS.SPY_REPORT);
      cc.setProtocol(FIPACONSTANTS.FIPA_SUBSCRIBE);
      cc.setConversationID(mf.getConversationID());
      Message m= new Message(cc);
      forward(m, mf.getHQ());
    }

    // Wait to ensure that the CANCEL messages are sent
    Thread.yield();
    try
    {
      Thread.sleep(500);
    }
    catch (Exception ex) {}

    DIAGNOSTICS.println("SPY:Spy shut down.", _owner, DIAGNOSTICS.LEVEL_MAX);

//    if(_sender!=null)
//      _sender.shutdown();
  }

  /**
   * Process the outgoing/incomming message: compare with the filter and check
   * the contents.
   * Returns true if the message should pass.
   */
  private boolean ProcessMessage(Message msg)
  {
    ACL acl= msg.getACL();

    if(_internalf.Match(acl))  // The message is for SPY process
    {
      boolean sendresponse= true;
      ACL re= new ACL();

      // Prepare response
      re.setPerformative(FIPACONSTANTS.INFORM);
      re.setSenderAID(_owner.getAID());
      re.setReceiverAID(acl.getSenderAID());
      re.setLanguage(SPYCONSTANTS.SPY_REPORT);
      re.setOntology(SPYCONSTANTS.SPY_LANGUAGE);
      re.setConversationID(acl.getConversationID());
      re.setInReplyTo(acl.getReplyWith());

      try
      {
        Action action;

        Object content= acl.getContentObject();

        // Type of content depend on message transport protocol used
        if(content instanceof Action)
          action= (Action)content;
        else
          action= new Action((String)acl.getContentObject());

        if(action.getAgent().getName().equalsIgnoreCase(_owner.getAID().getName()))
        {
          if(acl.getPerformative().equalsIgnoreCase(FIPACONSTANTS.REQUEST))
          {      // GET STATISTICS + CLEAR STATISTICS
            if(action.comparePredicate(SPYCONSTANTS.CLEAR))
            {
              if(action.compareSubject(SPYCONSTANTS.STATISTICS))
              {
                _instats.clear();
                _outstats.clear();
                re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
              }
              else
                throw new OntException("Invalid Subject to CLEAR: "+action.getSubject());
            }
            else if(action.comparePredicate(SPYCONSTANTS.GET))
            {
              if(action.compareSubject(SPYCONSTANTS.STATISTICS))
                re.setContentObject("("+SPYCONSTANTS.STATISTICS+" "+
                      SPYCONSTANTS.INCOMMING+" "+_instats+
                      SPYCONSTANTS.OUTGOING+" "+_outstats+")");
              else
                throw new OntException("Invalid Subject to GET: "+action.getSubject());
            }
            else
              throw new OntException("Invalid predicate for request: "+action.getPredicate());
          }
          else if(acl.getPerformative().equalsIgnoreCase(FIPACONSTANTS.CANCEL))
          {     // CANCEL FILTER SUBSCRIPTION
            if(action.comparePredicate(SPYCONSTANTS.CLEAR))
            {
              if(action.compareSubject(SPYCONSTANTS.FILTER))
              {
                _filter.remove(new MsgFilter.ConversationComparator(acl.getConversationID()));
//                re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
                sendresponse= false;
              }
              else
                throw new OntException("Invalid Subject to CLEAR: "+action.getSubject());
            }
            else
              throw new OntException("Only (clear (filter... is permitted for CANCEL performative");
          }
          else if(acl.getPerformative().equalsIgnoreCase(FIPACONSTANTS.SUBSCRIBE))
          {     // PUT FILTER
            if(action.comparePredicate(SPYCONSTANTS.PUT))
            {
              MsgFilter filter= new MsgFilter(action.getSubject());
              if(!_filter.contains(new MsgFilter.HQComparator(filter.getHQ().getName())))
              {
                filter.setConversationID(acl.getConversationID());
                _filter.add(filter);
                re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
              }
              else
                throw new OntException("Only one filter per HQ allowed.");
            }
            else
              throw new OntException("Only (put (filter... is premitted for SUBSCRIBE performative");
          }
          else
            throw new OntException("Unsupported performative: "+acl.getPerformative());

//          if(action.comparePredicate(SPYCONSTANTS.CLEAR))
//          {
//            if(action.compareSubject(SPYCONSTANTS.STATISTICS))
//            {
//              _instats.clear();
//              _outstats.clear();
//              re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
//            }
//            else if(action.compareSubject(SPYCONSTANTS.FILTER))
//            {
//              _filter= null;
//              re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
//            }
//            else
//              throw new OntException("Invalid Subject to CLEAR: "+action.getSubject());
//          }
//          else if(action.comparePredicate(SPYCONSTANTS.GET))
//          {
//            if(action.compareSubject(SPYCONSTANTS.STATISTICS))
//              re.setContentObject("("+SPYCONSTANTS.STATISTICS+" "+
//                    SPYCONSTANTS.INCOMMING+" "+_instats+
//                    SPYCONSTANTS.OUTGOING+" "+_outstats+")");
//            else
//              throw new OntException("Invalid Subject to GET: "+action.getSubject());
//          }
//          else if(action.comparePredicate(SPYCONSTANTS.PUT))
//          {
//            _filter= new MsgFilter(action.getSubject());
//            _filter.setConversationID(acl.getConversationID());
//            re.setContentObject("("+SPYCONSTANTS.DONE+" "+action+")");
//          }
//          else
//            throw new OntException("Invalid predicate: "+action.getPredicate());
//
//          re.setPerformative(FIPACONSTANTS.INFORM);
        }
        else
          throw new OntException("Action to be performed by someone else.");
      }
      catch(OntException e)
      {
        re.setPerformative(FIPACONSTANTS.NOT_UNDERSTOOD);
        re.setContentObject(acl);
        DIAGNOSTICS.println("SPY:Message not understood. Reason: "+e.getMessage(), _owner, DIAGNOSTICS.LEVEL_4);
      }

      if(sendresponse)
        forward(new Message(re), acl.getSenderAID());
      return false;
    }

    Iterator i= _filter.iterator();
    while(i.hasNext())
    {
      MsgFilter f= (MsgFilter)i.next();

      if(f.Match(acl))   // CC. this Msg to HQ
      {
        ACL cc= new ACL();

        cc.setPerformative(FIPACONSTANTS.INFORM);
        cc.setSenderAID(_owner.getAID());
        cc.setReceiverAID(f.getHQ());
        cc.setContentObject(msg.getACL());
        cc.setInReplyTo((String)msg.getEnvelope().getSlotValue(SPYCONSTANTS.SPY_TAG));
        cc.setLanguage(SPYCONSTANTS.SPY_LANGUAGE);
        cc.setOntology(SPYCONSTANTS.CAPTURED_MESSAGE);
        cc.setProtocol(FIPACONSTANTS.FIPA_SUBSCRIBE);
        cc.setConversationID(f.getConversationID());
        Message m= new Message(cc);
        forward(m, f.getHQ());
      }
    }
    return true;
  }

  private void forward(Message msg, AgentID receiver)
  {
    if(receiver.getName().equalsIgnoreCase(_owner.getAID().getName()))
    {
    // Ignore - FIPA-OS is unable to hangle such a message
//     _receiver.receiveMessage(msg);        //Mgs to self
    }
    else
      _sender.sendMessage(msg);             //Outgoing message
  }


  /**
   * Unique message ID
   */
  private  long _mID;

  /**
   * The agent that this is the spy of
   */
   private FIPAOSAgent _owner;

  /**
   * Link to the original MTS, that hanhles outgoing messages
   * and passes to us the incomming ones.
   */
  private MessageSender _sender;

  /**
   * Link to the original Conversation Manager, that handles incomming messages
   * and passesthe outgoing messages to us.
   */
  private MessageReceiver _receiver;

  /**
   * The vector of filters to compare the messages with.
   */
  private Vector _filter;

  /**
   * The filter to find out messages for the Spy
   */
   private final MsgFilter _internalf;

  /**
   * Statistics of incomming/outgoing messages.
   * The performative is the table key, the table entry is the actual count.
   */
  private MsgStats _instats;
  private MsgStats _outstats;

  public static void main( String args[] )
  {
//    try
//    {
//    }
//    catch(OntException e)
//    {
//      System.out.println(e.getMessage());
//    }

    return;
  }
}