//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//    Gerstner Laboratory, Czech Technical University in Prague,
//    Karlovo namesti 13, 121 35 Prague 2, Czech Republic. All Gerstner
//    Laboratory Modifications are Copyright (C) 2001 Gerstner Laboratory.
//    All Rights Reserved.
//
//******************************************************************************

/**
 * Title:        PingPong agent for MessageSpy testing
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Gerstner Laboratory, FEE-CTU, Prague
 * @author Michal Dobek, michaldobisek@bigfoot.com
 * @version 1.0
 */

package bp.pingpong;

//----------------------------------------------------------------------

// Java
import java.util.*;
import javax.swing.SwingUtilities;
import java.lang.Runnable;

// Import the fipa classes
import fipaos.ont.fipa.*;
import fipaos.parser.sl.SLParser;
import fipaos.parser.ParserException;

// Import the agent management classes (as we are registering with the platform)
import fipaos.ont.fipa.fipaman.*;

// We will also need to import agent classes
import fipaos.agent.*;
import fipaos.agent.task.*;
import fipaos.agent.conversation.*;
import fipaos.agent.conversation.content.*;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;

// Finally import the diagnostics class for output
import fipaos.util.DIAGNOSTICS;

// My imports
import bp.ont.Constructor;

//----------------------------------------------------------------------

/**
 * The Generic Agent is a tutorial agent showing how to construct an 'empty'
 * agent that simply registers with the local agent platform.
 * <p>Please read both the Javadoc comments and the code comments to follow
 * the tutorial.
 */
public class PingPongAgent extends FIPAOSAgent
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	// This is the type of the agent that we use when registering with the DF
	private static final String AGENT_TYPE	= "ping-pong-agent";


	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

  private PingPongGUI GUI = null;
  private IdleTask _idletask= null;

	//----------------------------------------------------------------------
	// CONSTRUCTOR
	//----------------------------------------------------------------------

	/**
	 * An agent constructor takes three parameters - the location of the platform
	 * profile file (usually fipa-os\profiles\platform.profile), the agent name
	 * (e.g. df, ams), and the owner of the agent (e.g. Nortel, Bob).
	 *
	 * @param platform_profile The location of the platform profile
	 * @param name The agent name
	 * @param owner The owner of this agent
	 */
	public PingPongAgent( String platform_profile, String name, String owner )
	{

	/* First you must initialise the agent with a call to the FIPAOSAgent
	 * constructor. The arguments are the same three arguments to this constructor
	 * plus some instantiation flags.
	 * Arg 4 specifies if the agent should use the Task Manager - this does not have to be true
	 * but if you want to use the automatic platform registration (as this agent does) then
	 * you should set this to true. Tasks also make agent construction easier for more
	 * complicated agents. A discussion of programming agents using states is in the
	 * tutorial but is not included in this agent.
	 */
		super( platform_profile, name, owner, true );

    _idletask= new IdleTask();
    setListenerTask( _idletask );

		// Now we can register with the local AMS (i.e. the platform) as the fist action our agent
		// makes
		try
		{
			// Attempt to register with AMS
			registerWithAMS();
			DIAGNOSTICS.println("Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch (AMSRegistrationException amsre)
		{
			// An exception has occured - this indicates that the AMS registration failed for some reason
			// Display the exception
			DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );

			// We can easily find the exception reason from the exception....
			String reason = amsre.getExceptionReason();

			// If the exception reason is not "already-registered", we can probably ignore the failure
			// (since this Agent was probably not de-registered the last time it ran!)
			if ( reason == null || !reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED))
			{
				// Shutdown nicely :)
				shutdown();
				return;
			}

		}

		// Now we can register with the local DF
		try
		{
			// Register with DF
			registerWithDF( AGENT_TYPE );
			DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch (DFRegistrationException dfre)
		{
			// An exception has occured - this indicates that the DF registration failed for some reason
			// Display the exception
			DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );

			// We can easily find the exception reason from the exception....
			String reason = dfre.getExceptionReason();

			if ( reason == null || !reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) )
			{
				// Shutdown nicely :)
				shutdown();
				return;
			}

		} // end catch

    try
    {
      GUI = new PingPongGUI(this);
      GUI.setVisible(true);
    }
		catch( Throwable t )
		{
      DIAGNOSTICS.println("Can't start GUI, quitiing: "+t, this, DIAGNOSTICS.LEVEL_MAX );
      // Shutdown nicely :)
      shutdown();
      return;
		}
	}


	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------




	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/**
	 * Shuts down the agent by first deregistering with the DF and AMS
	 * (if registered) and then invoking shutdown() on the FIPAOSAgent shell
	 */
	public synchronized void shutdown()
	{
		// Check if we've registered with the DF
		if (registeredWithDF() == true)
		{
			try
			{
				// Attempt to deregister with DF
				deregisterWithDF();
				DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
			}
			catch (DFRegistrationException dfre)
			{
				// A problem deregistering occured..... we can obtain the reason though!
				String reason = dfre.getExceptionReason();

				DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
			} // end catch
		}

		// Check if we're registered with the AMS
		if (registeredWithAMS() == true)
		{
			try
			{
				// Deregister with AMS
				deregisterWithAMS();
				DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
			}
			catch (AMSRegistrationException amsre)
			{
				// A problem deregistering occured..... we can obtain the reason though!
				String reason = amsre.getExceptionReason();

				DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
			}
		}

    if(GUI!=null)
    {
      GUI.dispose();
      GUI= null;
    }

		// Now call shutdown in the agent shell to release the core components
		super.shutdown();
	}

  public void donePingPongAgent_IdleTask(Task t)
  {
    shutdown();
    return;
  }

	public class IdleTask extends Task
	{
		public IdleTask()
		{
			// Don't do anything on instantiation
		}

    protected void StartTask()
    {
      // Don't do anything - we are just waiting for match requests
    }

    public void handleRequest( Conversation conv )
		{
			// Indicate what is going on
		 	DIAGNOSTICS.println( "PA: Recieved request. ", this, DIAGNOSTICS.LEVEL_4 );

      try
      {
        ACL acl = conv.getACL(conv.getLatestMessageIndex());

        Constructor msg= new Constructor(acl);

        if(msg.CompareOp("die"))
        {
          DIAGNOSTICS.println( "PA: Quitting.", this, DIAGNOSTICS.LEVEL_4 );
          done();
        }
        else if(msg.CompareOp("play"))
        {
          DIAGNOSTICS.println( "PA: Playing- Initiating", this, DIAGNOSTICS.LEVEL_4 );
          StartPingPong(conv, new AgentID(msg.GetStrParam(":opponent")), msg.GetIntParam(":count"));

          // Send Agree
          ACL resp = conv.getFilledInACL();
          resp.setPerformative(FIPACONSTANTS.AGREE);
          resp.setSenderAID(_owner.getAID());
          resp.setReceiverAID(acl.getSenderAID());
          resp.setContentObject(acl);
          forward(resp);
        }
        else if(msg.CompareOp("ping"))
        {
          DIAGNOSTICS.println( "PA: Playing- Answering", this, DIAGNOSTICS.LEVEL_4 );
          newTask( new PongTask( conv ), conv);
        }
        else
          _owner.sendNotUnderstood(acl, "unknown command");
      }
      catch(Throwable t)
      {
        DIAGNOSTICS.println( "PA: Error - "+t.getMessage(), this, DIAGNOSTICS.LEVEL_4 );
      }
		}

    public void donePingTask(Task t)
    {
      PingTask p= (PingTask)t;

      Conversation c= p.GetInitiator();   // Send DONE
      if(c!=null)
      {
        ACL resp = c.getFilledInACL();
        resp.setReceiverAID(p.GetInitiatorAID());
        resp.setPerformative(FIPACONSTANTS.INFORM);
        resp.setContentObject( p.GetResult() );
        forward( resp );
      }

      // And show the result in the table
      fireInsertResult(p.GetGameState(), p.GetOpponent().getName());
      return;
    }

    public void donePongTask(Task t)
    {
      PongTask p= (PongTask)t;
      fireInsertResult(p.GetGameState(), p.GetOpponent().getName());
      return;
    }

    synchronized public void StartPingPong(Conversation Initiator, AgentID Opponent, int n)
    {
      newTask(new PingTask(Initiator, Opponent, n));
    }

	}

  synchronized public void StartPingPong(Conversation Initiator, AgentID Opponent, int n)
  {
    _idletask.StartPingPong(Initiator, Opponent, n);
  }

  public void StartDFSearch()
  {
    DFAgentDescription df_desc = new DFAgentDescription();

    ServiceDescription sd = new ServiceDescription();
    sd.setServiceName(AGENT_TYPE);
    sd.setServiceType(AGENT_TYPE); 			// We want PingAgent agents
    df_desc.setAgentServices( new ServiceDescription[] { sd } );

    _tm.newTask(new DFSearchTask(df_desc));
  }

  public void doneDFSearchTask(Object o)
  {
    // Search finished, update the combo
    fireSearchResults((DFAgentDescription [])o);
  }

  public void activate()
  {
    if(GUI!=null)
      GUI.setVisible(true);
  }

  private void fireInsertResult(final String result, final String opponent)
  {
    // Ensure the GUI is modified only from AWT Event Dispatch Thread
    SwingUtilities.invokeLater(new Runnable() {
      public void run()
      { GUI.insertResult(result, opponent); }
    } );
  }

  private void fireSearchResults(final DFAgentDescription [] r)
  {
    // Ensure the GUI is modified only from AWT Event Dispatch Thread
    SwingUtilities.invokeLater(new Runnable() {
      public void run()
      { GUI.SearchResults(r); }
    } );
  }

	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------

  /**
   * A main method to enable the agent to be run from the command line.
   * The agent takes two command line parameters - the platform profile location,
   * which is set up by the SetupFIPAOS.bat file to be in the system variable %PLATFORM%,
   * and the agent name.
   */
	public static void main( String args[] )
	{

		// Check for the correct number of arguments
		if( args.length < 2 )
		{

			System.out.println( "Usage: bp.pingpong.PingPongAgent location-of-platform-profile agent-name" );

		}
		else
		{

      try
      {
        javax.swing.UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
      } catch (Exception e) { }

			// If number of args is correct then run the agent - in this case the ownership
			// is always "FIPA-OS"
			// args[ 0 ] = the platform profile location
			// args[ 1 ] = the agent name
			new PingPongAgent( args[ 0 ], "Ping", "MiDo" );
      new PingPongAgent( args[ 0 ], "Pong", "MiDo" );

		}

	}


} // End PingPongAgent