//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//    Gerstner Laboratory, Czech Technical University in Prague,
//    Karlovo namesti 13, 121 35 Prague 2, Czech Republic. All Gerstner
//    Laboratory Modifications are Copyright (C) 2001 Gerstner Laboratory.
//    All Rights Reserved. MessageSpy support addes. Modifications marked
//    with "MessageSpy" tag
//
//******************************************************************************

package fipaos.agent;

// -------------------------------------------------------------------------

import fipaos.ont.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.agent.conversation.*;
import fipaos.agent.conversation.standard.ConversationManagerImpl;
import fipaos.agent.task.*;
import fipaos.agent.profile.*;
import fipaos.platform.ams.*;
import fipaos.platform.df.*;
import fipaos.mts.*;
import fipaos.mts.service.*;
import fipaos.util.*;

import java.util.*;
import java.io.*;
import java.net.*;
import java.lang.*;

//START MessageSpy Code
import bp.meta.spy.MessageSpy;
//END MessageSpy Code

// -------------------------------------------------------------------------

/**
 * Base FIPA-OS agent class implementing FIPA-OS specifics for agent to
 * agent communication.
 *
 * Override <pre>notify()</pre> deal with updated Conversations directly
 * in the underlying Agent implementation, or use the <pre>setListenerTask()</pre>
 * to set a default Task if you wish to use the TaskManager to implement the Agent.
 * <br>
 * <br>
 */
public abstract class FIPAOSAgent implements ConversationListener
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	private static final String IDLE = "IDLE";

	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

	/**
	 * MessageSender to deal with our outgoing messages
	 */
	private MessageSender _sender = null;

	/**
	 * Is agent ready to accept another pushed message (push model only)
	 */
	private boolean _ready_to_accept;

	/**
	 * Lock used to prevent messages being pushed before the Agent is
	 * ready for them
	 */
	private Object _push_lock = new Object();

	/**
	 * Task manager.
	 */
	protected TaskManager _tm;

	/**
	 *  Planning and Scheduling engine.
	 */
	//	protected PlannerScheduler _ps;

	/**
	 * Conversation manager.
	 */
	private ConversationManager _cm;

  // START MessageSpy Code
  /**
   * MessageSpy
   */
   private MessageSpy _spy;
   // END MessageSpy Code

	/**
	 * MTS for this Agent
	 */
	private MTS _mts;

	/**
	 * My AgentID
	 */
	private AgentID _aid;

	/**
	 * My profile
	 */
	private AgentProfile _profile;

	/**
	 * Platform profile
	 */
	private PlatformProfile _platform_profile;

	/**
	 * Indicates if the Agent has shutdown
	 */
	private boolean _shutdown;

	/**
	 * Database Profile
	 */
	protected DatabaseProfile _database_profile;

	/**
	 * Ownership info
	 */
	private String _owner;

	/**
	 * Task to handle top-level task completions (i.e. pass callbacks to
	 * Agent implementation!)
	 */
	private TopLevelHandlerTask _top_level_handler;

	/**
	 * Agent states: unknown, waiting, suspended, transit, initiated, active
	 */
	public String _state = FIPAMANCONSTANTS.UNKNOWN;

	/**
	 * List of Agents that this Agent is registered with
	 */
	public List _registered_with = Collections.synchronizedList( new LinkedList() );

	//----------------------------------------------------------------------
	// CONSTRUCTORS
	//----------------------------------------------------------------------

	/**
	 * Constructor method to initialise an Agent.  This loads the AgentProfile
	 * and creates an AgentComms instance for the Agent with the required
	 * CommsTransports
	 *
	 * @deprecated Other constructors are now preferred, as they allow use of
	 *             conversation manager and task manager.
	 * @param platform_profile_location The location of the platform.profile (a fully qualified file name)
	 * @param agent_name globally unique identifier for the agent
	 * @param ownership the ownership of the agent
	 */
	private FIPAOSAgent( String platform_profile_location,
	String agent_name )
	{
		// Attempt to load platform profile
		try
		{
			_platform_profile = (PlatformProfile)XMLDataBinding.unmarshallFile(platform_profile_location);
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "An error occured whilst loading the platform profile: " + t,
			this );
			throw new RuntimeException( "An error occured whilst loading the platform profile: " + t );
		}


		DIAGNOSTICS.println( "Attempting to load AgentProfile now....", this );


		// Attempt to load the Agent's profile
		try
		{
			// Work out filename!
			String location = _platform_profile.getProfileDirectory();
			if( !location.endsWith( "/" ) && !location.endsWith( "\\" ) )
			{
				location += "/";
			}
			String ap_location = location + agent_name + ".profile";
			String dp_location = location + "default.profile";

			// Get profile(s)
			AgentProfile dp = null;
			AgentProfile ap = null;

			try
			{
				// Load default profile, if available
				dp = (AgentProfile)XMLDataBinding.unmarshallFile(dp_location);
			}
			catch( Throwable t )
			{
				DIAGNOSTICS.println( "Couldn't find default Agent profile" );
			}

			try
			{
				// Load agent specific profile, if available
				ap = (AgentProfile)XMLDataBinding.unmarshallFile(ap_location);
			}
			catch( Throwable t )
			{
				if( _profile != null )
				{
					DIAGNOSTICS.println( "Couldn't find specific Agent profile, using default profile" );
				}
			}

			// No profile loaded at all?
			if( ap == null && dp == null )
			{
				throw new RuntimeException( "An error occured whilst loading the agents profile - neither " +
					"default profile or Agent specific profile have been loaded!" );
			}

			// Default loaded OK?
			if( dp != null )
			{
				// Yes, make it default
				_profile = dp;

				// Override default setting from AP, if available
				if( ap != null )
				{
					// Override database setup
					DatabaseProfile dbp = ap.getDatabaseProfile();
					DatabaseProfile dbp2 = _profile.getDatabaseProfile();
					if( dbp != null )
					{
						if( dbp.getDatabaseLocation() != null )
						{
							dbp2.setDatabaseLocation( dbp.getDatabaseLocation() );
						}

						if( dbp.getDatabaseType() != null )
						{
							dbp2.setDatabaseType( dbp.getDatabaseType() );
						}
					}

					if( ap.getInternalAddresss() != null && !ap.getInternalAddresss().isEmpty() )
					{
						_profile.getInternalAddresss().addAll( ap.getInternalAddresss() );
					}

					// Override protocol setup
					if( ap.getProtocolProfile() != null )
					{
						_profile.getProtocolProfile().getProtocolMappings().addAll( ap.getProtocolProfile().getProtocolMappings() );
					}
				}
			}
			else
			{
				// Ap only
				_profile = ap;
			}
		}
		catch( RuntimeException e )
		{
			throw e;
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "An error occured whilst loading the agents profile: ", t,
			this );
			throw new RuntimeException( "An error occured whilst loading the agents profile: " + t );
		}

		try
		{
			// Create Agent's ID
			_aid = new AgentID();
			_aid.setName( agent_name + "@" + _platform_profile.getHAPName() );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error whilst generating AgentID: " + t );
			throw new RuntimeException( "Error whilst generating AgentID: " + t );
		}
	}


	/**
	 * Constructor method to initialise an Agent. This loads the AgentProfile
	 * and creates an AgentComms instance for the Agent with the required
	 * CommsTransports. This constructor creates an agent that uses the Task Manager
	 * by default. This is going to be the required way of creating an agent in
	 * future versions of FIPA-OS, so this constructor should provide forward
	 * compatability. It will initialise the agent to use the Conversation Manager
	 * as well (as this is required for use of the Task Manager).
	 *
	 * @param platform_profile_location The location of the platform.profile (a fully qualified file name)
	 * @param agent_name globally unique identifier for the agent
	 * @param ownership the ownership of the agent
	 */
	public FIPAOSAgent( String platform_profile_location, String agent_name,
	String ownership )
	{
		this( platform_profile_location, agent_name, ownership, true );
	}


	/**
	 * Constructor method to initialise an Agent.  This loads the AgentProfile
	 * and creates an AgentComms instance for the Agent with the required
	 * CommsTransports. Can optionally create a Task Manager in the agent. If you are
	 * using the task manager then you must also call setListenerTask( Task ) in your
	 * constructor code in order to set up the default task. This default task handles
	 * all incoming messages that are new (i.e. that are not replies to message that we
	 * have previously sent). If you do not register a default task then incoming
	 * messages will be ignored.
	 *
	 * @param platform_profile_location The location of the platform.profile (a fully qualified file name)
	 * @param agent_name unique name of the Agent within its platform
	 * @param ownership the ownership of the agent
	 * @param using_task_manager Selected if a task manager is to be used
	 */
	public FIPAOSAgent( String platform_profile_location,
	String agent_name,
	String ownership,
	boolean using_task_manager )
	{
		this( platform_profile_location, agent_name );

		// Save owner info.
		_owner = ownership;

		// Agent initially unready for a push
		_ready_to_accept = false;

		// Get the DB class name to use
		try
		{
			_database_profile = _profile.getDatabaseProfile();
		}
		catch( Throwable t )
		{
		}

		// Now setup the core-components of the Agent
		try
		{
			// Setup the MTS
			_mts = new MTS( new fipaos.util.URL(_platform_profile.getAMSAddress()), _profile.getInternalAddresss(), _aid );

			// Default conversation listener is this
			ConversationListener _cm_target = this;

			// Check if we're using the TaskManager
			if( using_task_manager )
			{
				_top_level_handler = new TopLevelHandlerTask();
				_tm = new TaskManager( this, null, _top_level_handler );
				_cm_target = _tm;

				// Setup a default TM Listener for the time being
				// TODO: This should be the PlannerScheduler
				new TaskManager.DefaultTaskManagerListener( _tm );
			}

			// Setup the PlannerScheduler
			// TODO: Use this for real
			// _ps = new PlannerScheduler(this, _tm);

			// Setup the ConversationManagerImpl
			_cm = new ConversationManagerImpl( _cm_target, _mts, this, _database_profile, _profile.getProtocolProfile() );

			// Save a reference to the default MessageSender
			_sender = _cm;

      // START MessageSpy Code
      // Create the SPY and link it with _MTS and _cm
      _spy= new MessageSpy(this);
      _spy.setMessageReceiver(_cm);
      _spy.setMessageSender(_mts);
      _mts.setMessageReceiver(_spy);
      _cm.setMessageSender(_spy);
      // END MessageSpy Code

			DIAGNOSTICS.println( "Agent has been assigned AgentID " + _aid, this, DIAGNOSTICS.LEVEL_MAX );
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "FATAL ============================================\n" +
			"Something went wrong whilst starting up the Agent:\n",
			t, this, DIAGNOSTICS.LEVEL_MAX );
		}
	}

	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/**
	 * Forward an acl message out via the ConversationManagerImpl
	 *
	 * @param acl ACL containing an agent communication language
	 * (acl) message
	 */
	public void forward( ACL acl )
	{
		Message msg = new Message( acl );

		// Add time-stamp
		msg.getEnvelope().setSlotValue( Service.FIPAOS_DEBUG + "as-send-time",
																				new Long( System.currentTimeMillis() ) );

		_sender.sendMessage( msg );
	}


	/** Blocks until the Agent is ready to receive messages/conversations
	 */
	public void waitForPushLock()
	{
		// Return straight away if its OK
		if( _ready_to_accept )
		{
			return;
		}

		// Check that its OK to push (blocks if it isn't)
		synchronized( _push_lock )
		{
			while( !_ready_to_accept )
			{
				try
				{
					_push_lock.wait();
				}
				catch( InterruptedException e )
				{
				}
			}
		}
	}



	/**
	 * Shutdown the Agent safely
	 */
	public synchronized void shutdown()
	{
		DIAGNOSTICS.println( "Agent " + _aid.getName() + " is attempting to shutdown...",
		DIAGNOSTICS.LEVEL_MAX );

		// Kill off the CM, TM and MTS, if started
		if( _cm != null )
		{
			_cm.shutdown();
			_cm = null;
		}

		if( _tm != null )
		{
			_tm.shutdown();
			_tm = null;
		}

    // START MessageSpy Code
    if( _spy != null )
    {
      _spy.shutdown();
      _spy= null;
    }
    // END MessageSpy Code

		if( _mts != null )
		{
			_mts.shutdown();
		}


		_state="unknown";
		_shutdown = true;

		DIAGNOSTICS.println( "Agent " + _aid.getName() + " has shutdown",
		DIAGNOSTICS.LEVEL_MAX );
	}

	/**
	 * Indicates if this Agent has been shutdown
	 */
	public boolean isShutdown()
	{
		return _shutdown;
	}

  /**
   * Returns the ownership id of the agent.
   */
	public String getOwnership()
	{
		return _owner;
	}


	/**
	 * Gets the profile for this Agent
	 */
	public AgentProfile getProfile()
	{
		return _profile;
	}


	/**
	 * Gets the platform profile for this Agent
	 */
	public PlatformProfile getPlatformProfile()
	{
		return _platform_profile;
	}


	/**
	 * Gets the AgentID of this Agent
	 */
	public AgentID getAID()
	{
		return (AgentID) _aid.clone();
	}


	/**
	 * Gets the local HAP name
	 */
	public String getHAP()
	{
		return _platform_profile.getHAPName();
	}


	/**
	 * This method should be overriden by the FIPAOSAgent - it is invoked when the
	 * user of the local machine wishes it to do something (usually display a GUI)
	 * <p>By default it does nothing
	 */
	public void activate()
	{
	}

	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------

	/**
	 * Method to retreive a conversation from the active list
	 *
	 * @param conv The conversation id of the Conversation object to retreive
	 */
	public Conversation getCurrentConversation( String conv ) throws InvalidConversationIDException
	{
		return _cm.getConversation( conv );
	}


	/**
	 * Returns the agent AID of the local platform AMS.
	 *
	 * @return The local AMS GUID
	 */
	public String getLocalAMS()
	{
		return getLocalAMSAID().toString();
	}


	/**
	 * Returns the agent AID of the local platform AMS.
	 *
	 * @return The local AMS GUID
	 */
	public AgentID getLocalAMSAID()
	{
		// Create empty AID, add default AMS name + address
		AgentID aid = new AgentID();
		aid.setName( FIPACONSTANTS.AMS + BASETOKENS.AT + _platform_profile.getHAPName() );
		aid.setAddresses( new LinkedList(
		Collections.singleton( (Object) new fipaos.util.URL(_platform_profile.getAMSAddress()) ) ) );

		return aid;
	}


	/**
	 * Returns the agent AID of the local platform DF.
	 *
	 * @return The local DF GUID
	 */
	public String getLocalDF()
	{
		return getLocalDFAID().toString();
	}


	/**
	 * Returns the agent AID of the local platform DF.
	 *
	 * @return The local DF GUID
	 */
	public AgentID getLocalDFAID()
	{
		// Create empty AID, add default AMS name + address
		AgentID aid = new AgentID();
		aid.setName( FIPACONSTANTS.DF + BASETOKENS.AT + _platform_profile.getHAPName() );

		return aid;
	}

	/**
	 *	Returns the state of the agent
	 *	Currently, only 2 states might be returned: "unknown" and "active"
	 **/

	public String getState()
	{
		return _state;
	}


	/**
	 * Sets the agent's idle task - this method MUST be used if you are using the task
	 * manager as it defines which task object should handle incoming messages that we
	 * have not previously seen.
	 *
	 * @param idle_task The task object that should be used when the agent is idling
	 *        e.g. the task object that handles incoming messages that are not replies
	 * @return True if the IDLE task was set, false otherwise
	 **/
	protected boolean setListenerTask( Task idle_task )
	{

		if ( _tm != null )
		{

			// Add IDLE task
			_tm.newTask( idle_task, IDLE );

			// Messages can now be pushed if everything went OK
			startPushing();

			return true;
		}

		return false;

	}


	/**
	 * When using the push model or ConversationManagement, this method should be invoked
	 * in order to allow pushing to start
	 */
	protected void startPushing()
	{
		synchronized( _push_lock )
		{
			_ready_to_accept = true;
			_push_lock.notify();
		}

		_state="active";
	}


	/**
	 * Adds a newly downloaded protocol definition to the conversation manager to make it
	 * available for use.
	 *
	 * @param protocol_name The name of the protocol to add (e.g. fipa-new-protocol)
	 * @param protocol_defn The RDF definition of the protocol
	 * @return True if the protocol was added to the CM, false otherwise
	 **/
	protected boolean addNewProtocol( String protocol_name, String protocol_defn )
	{
		return _cm.addProtocol( protocol_name, protocol_defn );
	}


	/**
	 * Indicates if the agent is aware of the protocol specified.
	 *
	 * @param protocol_name The name of the protocol (e.g. fipa-new-protocol)
	 * @return True if the agents knows the protocol and can use it without throwing
	 * an exception, false otherwise
	 **/
	protected boolean knowsProtocol( String protocol_name )
	{
		return _cm.hasProtocol( protocol_name );
	}


	/**
	 * Method to get a new Conversation object from the conversation manager given
	 * the protocol specified
	 */
	public Conversation getNewConversation( String protocol )
	throws fipaos.agent.conversation.UnknownProtocolException
	{
		return _cm.addConversation( protocol );
	}


	/**
	 * Provides a reference to this Agent's ConversationManager implementation
	 */
	protected ConversationManager getCM()
	{
		return _cm;
	}

	//----------------------------------------------------------------------
	// PUBLIC METHODS - ConversationListener interface
	//----------------------------------------------------------------------

	/**
	 * Conversation listener method to deal with conversations being sent by the
	 * planner scheduler.  These should be overriden as default by the sub-class
	 * if necessary
	 *
	 * @param conv The updated conversation to deal with
	 */
	public void notify( Conversation conv )
	{
		fipaos.util.DIAGNOSTICS.println( "Received conversation update - sub class doesn't override notify() " +
		"and hence the update is not dealt with: \n" + conv.toString() );
		return;
	}

	/**
	 * Deals with multiple conversations at once - just calls notify( Conversation )
	 * for each one.  Override in sub class to replace this behavior. ( NOTE: When
	 * overriding this method, it should be noted that care should be taken because
	 * this method could be invoked at any time after the FIPAOSAgent()  constructor
	 * had been used.  Normally this method doesn't do anything until after startPushing()
	 * is invoked).
	 *
	 * @param enum An enumeration of updated conversations to deal with
	 */
	public void notify( Enumeration enum )
	{
		// Check that its OK to push (blocks if it isn't)
		waitForPushLock();

		while( enum.hasMoreElements() )
		{
			notify( (Conversation) enum.nextElement() );
		}
	}

	/**
	 * Sets the MessageSender for this
	 */
	public void setMessageSender( MessageSender ms )
	{
		_sender = ms;
	}


	/**
	 * Send a not understood message in response to the given ACL message
	 *
	 * @param acl The ACL object to reply to
	 **/
	public boolean sendNotUnderstood( ACL orig_acl )
	{
		return sendNotUnderstood( orig_acl, "internal-failure unknown" );
	}


	/**
	 * Send a not understood message in response to the given ACL message
	 *
	 * @param acl The ACL object to reply to
	 * @param reason The reason for not-understanding
	 **/
	public boolean sendNotUnderstood( ACL orig_acl, String reason )
	{
		boolean done = false;

		try
		{
			// Check that the given ACL isn't a not-understood
			if ( !orig_acl.getPerformative().equals( FIPACONSTANTS.NOT_UNDERSTOOD ) )
			{
				// Duplicate ACL
				ACL acl = (ACL) orig_acl.clone();

				// Setup message
				acl.setPerformative( FIPACONSTANTS.NOT_UNDERSTOOD );
				acl.setReceiverAID( acl.getSenderAID() );
				acl.setSenderAID( getAID() );
				acl.setInReplyTo( acl.getReplyWith() );
				acl.setReplyWith( null );

				// Set content
				acl.setContentObject( "(" + orig_acl.toString() + "(" + reason + "))" );

				// Attempt to send the not-understood
				forward( acl );

				// We're done
				done = true;
			}
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error occured when sending not-understood", t, this,
			DIAGNOSTICS.LEVEL_MAX );
		}

		return done;
	}


	//----------------------------------------------------------------------
	// PROTECTED/PUBLIC METHODS - Auto platform reg interface
	//----------------------------------------------------------------------

	/**
	 * Indicates if this agent is registered with the platform agent specified.
	 * This does not have to be a local platform agent. To test registration task
	 * of the local platform agents you should use the <code>registeredWithX()</code>
	 * methods of this class.
	 *
	 * @return True if this agent is registered with the platform agent specified, false otherwise
	 * @deprecated Please use the registeredWith( AgentID ) method
	 */
	protected boolean registeredWith( String agent )
	{
		try
		{
			return _registered_with.contains( new AgentID( agent ) );
		}
		catch( Throwable t )
		{
			AgentID aid = new AgentID();
			aid.setName( agent );
			return _registered_with.contains( aid );
		}
	}

	/**
	 * Indicates if this agent is registered with the platform agent specified.
	 * This does not have to be a local platform agent. To test registration task
	 * of the local platform agents you should use the <code>registeredWithX()</code>
	 * methods of this class.
	 *
	 * @return True if this agent is registered with the platform agent specified, false otherwise
	 */
	protected boolean registeredWith( AgentID agent )
	{
		return _registered_with.contains( agent );
	}

	/**
	 * Indicates if this agent is registered with the local AMS.
	 *
	 * @return True if this agent is registered with the local AMS, false otherwise
	 */
	protected boolean registeredWithAMS()
	{
		return registeredWith( getLocalAMSAID() );
	}

	/**
	 * Indicates if this agent is registered with the local DF.
	 *
	 * @return True if this agent is registered with the local DF, false otherwise
	 */
	protected boolean registeredWithDF()
	{
		return registeredWith( getLocalDFAID() );
	}

	/**
	 * Registers this agent with the local platform AMS. The platform state of the agent
	 * is automatically set to ACTIVE.
	 */
	protected void registerWithAMS() throws AMSRegistrationException
	{

		registerWithAMS(new AMSRegistrationTask());

	} // end registerWithAMS()

	/**
	 * Registers this agent with the AMS specified. The platform state of the agent
	 * is automatically set to ACTIVE.
	 *
	 * @param ams The AID of the AMS to register with
	 */
	protected void registerWithAMS( AgentID ams ) throws AMSRegistrationException
	{

		registerWithAMS(new AMSRegistrationTask( (AgentID) ams.clone() ));

	} // end registerWithAMS()

	/**
	 * Registers this agent with the AMS specified. The platform state of the agent
	 * is set to the state specified.
	 *
	 * @param ams The AID of the AMS to register with
	 * @param state The platform state of the agent (e.g. ACTIVE - see FIPA spec for details)
	 */
	protected void registerWithAMS( AgentID ams, String state ) throws AMSRegistrationException
	{

		registerWithAMS(new AMSRegistrationTask( (AgentID) ams.clone(), state ));

	} // end registerWithAMS()

	/**
	 * Registers this agent using the AMS registration task specified.
	 *
	 * @param ams_task The AMSRegistrationTask to register with
	 */
	protected void registerWithAMS(AMSRegistrationTask ams_task) throws AMSRegistrationException
	{
		Object result = null;

		try
		{
			result = SynchronousTask.executeTask(_tm, ams_task);
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception when starting a Task", t, DIAGNOSTICS.LEVEL_MAX );
		}

		if (result != null)
		{
			if (result instanceof Object[])
			{
				Object failure[] = (Object[])result;

				AgentID agent = (AgentID)failure[0];
				String reason = (String)failure[1];

				if (reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED))
				{
					_registered_with.add(agent);
				}

				String msg = "Registration with " + agent + " failed :" + reason;
				throw new AMSRegistrationException(msg, reason);
			}
			else if (! (result instanceof AgentID))
			{
				throw new AMSRegistrationException("Registration with AMS failed");
			}
			else // normal completion, result = AgentID of AMS
			{
				_registered_with.add( (AgentID) ((AgentID)result).clone() );
			}
		}
		else
		{
			throw new AMSRegistrationException("Registration with AMS failed");
		}

	} // end registerWithAMS()

	/**
	 * Registers this agent with the local platform DF. The DF state of the agent
	 * is automatically set to ACTIVE. The agent type is set to that specified.
	 *
	 * @param agent_type The type of this agent, e.g. fipa-agent, fipa-arb etc
	 */
	protected void registerWithDF( String agent_type ) throws DFRegistrationException
	{

		registerWithDF(new DFRegistrationTask(agent_type));

	} // end registerWithDF()

	/**
	 * Registers this agent with the DF specified. The DF details of the agent are
	 * extracted from the DF Description object specified.
	 *
	 * @param df The AID of the AMS to register with
	 * @param df_desc A DF description that describes this agent
	 */
	protected void registerWithDF( AgentID df, DFAgentDescription df_desc ) throws DFRegistrationException
	{

		registerWithDF(new DFRegistrationTask(df, df_desc));

	} // end registerWithDF()

	/**
	 * Registers this agent with the DF specified. The DF state of the agent
	 * is automatically set to ACTIVE. The agent type is set to that specified.
	 *
	 * @param df The GUID/AID of the DF to register with
	 * @param agent_type The type of this agent, e.g. fipa-agent, fipa-arb etc
	 */
	protected void registerWithDF( AgentID df, String agent_type ) throws DFRegistrationException
	{

		registerWithDF(new DFRegistrationTask(df, agent_type));

	} // end registerWithDF()

	/**
	 * Registers this agent using the DF registration task specified.
	 *
	 * @param df_task The ADFRegistrationTask to register with
	 */
	protected void registerWithDF(DFRegistrationTask df_task) throws DFRegistrationException
	{
		Object result = null;

		try
		{
			result = SynchronousTask.executeTask(_tm, df_task);
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception when starting a Task", t, DIAGNOSTICS.LEVEL_MAX );
		}

		if (result != null)
		{
			if (result instanceof Object[])
			{
				Object failure[] = (Object[])result;

				AgentID agent = (AgentID)failure[0];
				String reason = (String)failure[1];

				if (reason.equals(FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED))
				{
					_registered_with.add(agent);
				}

				String msg = "Registration with " + agent + " failed :" + reason;
				throw new DFRegistrationException(msg, reason);
			}
			else if (! (result instanceof AgentID))
			{
				throw new DFRegistrationException("Registration with DF failed");
			}
			else // normal completion, result = AgentID of DF
			{
				_registered_with.add( (AgentID) ((AgentID)result).clone() );
			}
		}
		else
		{
			throw new DFRegistrationException("Registration with DF failed");
		}

	} // end registerWithDF()

	/**
	 * Deregisters this agent with the DF.
	 */
	protected void deregisterWithDF() throws DFRegistrationException
	{

		Object result = null;

		try
		{
			result = SynchronousTask.executeTask(_tm, new DFDeregistrationTask());
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception when starting a Task", t, DIAGNOSTICS.LEVEL_MAX );
		}

		if (result != null)
		{
			if (result instanceof Object[])
			{
				Object failure[] = (Object[])result;

				AgentID agent = (AgentID)failure[0];
				String reason = (String)failure[1];

				String msg = "Deregistration with " + agent + " failed :" + reason;
				throw new DFRegistrationException(msg, reason);
			}
			else if (! (result instanceof AgentID))
			{
				throw new DFRegistrationException("Deregistration with DF failed");
			}
		}
		else
		{
			throw new DFRegistrationException("Deregistration with DF failed");
		}

	} // end deregisterWithDF()

	/**
	 * Deregisters this agent with the DF.
	 */
	protected void deregisterWithAMS() throws AMSRegistrationException
	{

		Object result = null;

		try
		{
			result = SynchronousTask.executeTask(_tm, new AMSDeregistrationTask());
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Exception when starting a Task", t, DIAGNOSTICS.LEVEL_MAX );
		}

		if (result != null)
		{
			if (result instanceof Object[])
			{
				Object failure[] = (Object[])result;

				AgentID agent = (AgentID)failure[0];
				String reason = (String)failure[1];

				String msg = "Deregistration with " + agent + " failed :" + reason;
				throw new AMSRegistrationException(msg, reason);
			}
			else if (! (result instanceof AgentID))
			{
				throw new AMSRegistrationException("Deregistration with AMS failed");
			}
		}
		else
		{
			throw new AMSRegistrationException("Deregistration with AMS failed");
		}

	} // end deregisterWithAMS()

	//----------------------------------------------------------------------
	// PUBLIC METHODS - Legacy methods (from Agent class)
	//----------------------------------------------------------------------

	/** Previously returned the GUID of the agent.
	 * @deprecated The GUID is no longer in use - this now returns the AgentID
	 * associated with this Agent
	 */
	public String getGUID()
	{
		return _aid.toString();
	}

	/** Previously set the Agent's GUID
	 * @param new_guid
	 * @deprecated GUID's are no longer supported
	 */
	public void setGUID(String new_guid)
	{
		DIAGNOSTICS.println( "The agent has NOT been renamed" );
	}


	//----------------------------------------------------------------------
	// INNER-CLASSES
	//----------------------------------------------------------------------

	public class TopLevelHandlerTask extends Task
	{
	}

	//----------------------------------------------------------------------
	// TEST HARNESS METHODS
	//----------------------------------------------------------------------

	public static void main( String args[] )
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_2 );

		// Check if a platform profile has been specified
		if( args.length < 3 )
		{
			System.out.println( "Platform profile location is required as an argument!" );
			return;
		}

		System.out.println( "Starting up test agent....." );

		// Create test Agent
		new TestAgent( args[0], args[1], args[2] );
	}


	public static class TestAgent extends FIPAOSAgent
	{
		public TestAgent( String platform, String name, String target )
		{
			// Initialise Agent
			super( platform, name, "test-harness" );

			// Setup listener task
			this.setListenerTask( new IdleTask( target, this ) );

			//			this.registerWithAMS();
			//			this.registerWithDF( "test-agent" );
		}


		public class IdleTask extends Task
		{
			private AgentID _target;
			public FIPAOSAgent _parent;

			/**
			 * Constructor
			 */
			public IdleTask( String target, FIPAOSAgent parent )
			{
				// Save reference to parent
				_parent = parent;

				// Create AID for target Agent
				_target = new AgentID();
				if( target.indexOf( '@' ) == -1 )
				{
					_target.setName( target + '@' + _parent.getHAP() );
				}
				else
				{
					_target.setName( target );
				}
			}


			/**
			 * Invoked when the task is first started
			 */
			public void startTask()
			{
				// Start a WaitTask, to alert us every 5 seconds
				newTask( new WaitTask( 5000 ) );
			}

			/**
			 * Invoked when a message is received
			 */
			public void handleInform( Conversation conv )
			{
				// Get next message and display
				ACL acl = conv.getACL( conv.getLatestMessageIndex() );

				DIAGNOSTICS.println( "Received message: " + acl );
			}

			/**
			 * Invoked when its time to send a new message
			 */
			public void doneWaitTask( Task t )
			{
				try
				{
					// Send a message to the other Agent
					ACL msg = new ACL();

					msg.setPerformative( "inform" );
					msg.setSenderAID( _parent.getAID() );
					msg.setReceiverAIDs( Arrays.asList( new AgentID[]
					{ _target } ) );
					msg.setContentObject( "(Hello World!)" );

					System.out.println( "Sending message to " + _target );
					this.forward( msg );
				}
				catch( Throwable e )
				{
					// Error - display the exception
					e.printStackTrace();
				}

				// Start new WaitTask
				newTask( new WaitTask( 5000 ) );
			}


			public void DFSearchResults( DFAgentDescription descs[] )
			{
				DIAGNOSTICS.println( "Received DF results: " + Arrays.asList( descs ), this,
				DIAGNOSTICS.LEVEL_3 );
			}
		}
	}
}
